import numpy as np
import pandas as pd
import pytest

import pyranges1 as pr


def _base_frame_with_dtype(dtype: np.dtype) -> pd.DataFrame:
    return pd.DataFrame(
        {
            "Chromosome": pd.Series(["chr1", "chr1", "chr1", "chr1"], dtype="category"),
            "Start": pd.Series([1, 5, 20, 21], dtype=dtype),
            "End": pd.Series([10, 8, 25, 30], dtype=dtype),
            "Strand": pd.Series(["+", "+", "+", "+"], dtype="category"),
        },
    )


def test_cluster_overlaps_supports_uint32_issue_case() -> None:
    # Regression test for the historical unsupported dtype pair:
    # group dtype uint32 + position dtype uint32.
    gr = pr.PyRanges(_base_frame_with_dtype(np.uint32))

    out = gr.cluster_overlaps()

    assert out["Start"].dtype == np.dtype(np.uint32)
    assert out["End"].dtype == np.dtype(np.uint32)
    assert out["Cluster"].dtype == np.dtype(np.uint32)
    assert out["Cluster"].tolist() == [0, 0, 1, 1]


@pytest.mark.parametrize("dtype", [np.int8, np.uint8, np.int16, np.uint16, np.int32, np.uint32, np.int64])
def test_core_ops_support_common_integer_position_dtypes(dtype: np.dtype) -> None:
    gr = pr.PyRanges(_base_frame_with_dtype(dtype))

    clustered = gr.cluster_overlaps()
    merged = gr.merge_overlaps()

    assert clustered["Start"].dtype == np.dtype(dtype)
    assert clustered["End"].dtype == np.dtype(dtype)
    assert merged["Start"].dtype == np.dtype(dtype)
    assert merged["End"].dtype == np.dtype(dtype)
    assert clustered["Cluster"].dtype == np.dtype(np.uint32)


def test_uint64_outside_int64_range_raises() -> None:
    starts = np.array([2**63 + 1, 2**63 + 10], dtype=np.uint64)
    ends = starts + np.uint64(1)

    gr = pr.PyRanges(
        pd.DataFrame(
            {
                "Chromosome": ["chr1", "chr1"],
                "Start": starts,
                "End": ends,
            },
        ),
    )

    with pytest.raises(OverflowError, match="int64 range"):
        gr.cluster_overlaps()



from __future__ import annotations

import re
from importlib.metadata import PackageNotFoundError, version
from typing import Any

MIN_RURANGES_VERSION = "0.1.1"


def _version_tuple(v: str) -> tuple[int, ...]:
    nums = [int(x) for x in re.findall(r"\d+", v)]
    return tuple(nums) if nums else (0,)


def require_ruranges() -> Any:
    import ruranges

    try:
        installed = version("ruranges")
    except PackageNotFoundError as exc:
        msg = "pyranges1 requires the `ruranges` package."
        raise ImportError(msg) from exc

    if _version_tuple(installed) < _version_tuple(MIN_RURANGES_VERSION):
        msg = f"pyranges1 requires ruranges>={MIN_RURANGES_VERSION}; found {installed}."
        raise ImportError(msg)

    if not hasattr(ruranges, "numpy"):
        msg = (
            "Installed `ruranges` is missing the `ruranges.numpy` namespace. "
            f"Please install ruranges>={MIN_RURANGES_VERSION}."
        )
        raise ImportError(msg)

    return ruranges



from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

if TYPE_CHECKING:
    import contextlib

    from pyranges1 import PyRanges

    with contextlib.suppress(ImportError):
        from pyrle import PyRles  # type: ignore[import-not-found]


def _to_rle(
    ranges: "PyRanges",
    value_col: str | None = None,
    *,
    strand: bool = True,
    rpm: bool = False,
    **kwargs,
) -> "PyRles":
    try:
        from pyrle import PyRles  # type: ignore[import]
        from pyrle.methods import Rle, _coverage  # type: ignore[import]
    except ImportError as e:
        msg = "Using the coverage method requires that pyrle is installed."
        raise ImportError(msg) from e

    def _coverage_with_writable_arrays(df: pd.DataFrame, *, value_col: str | None = None) -> "Rle":
        values = df[value_col].astype(np.float64).to_numpy(copy=True) if value_col else np.ones(len(df))

        starts_df = pd.DataFrame({"Position": df.Start, "Value": values})[["Position", "Value"]]
        ends_df = pd.DataFrame({"Position": df.End, "Value": -1 * values})[["Position", "Value"]]
        coverage_df = pd.concat([starts_df, ends_df], ignore_index=True)
        coverage_df = coverage_df.sort_values("Position", kind="mergesort")
        coverage_df.loc[:, "Position"] = coverage_df.Position.astype(np.int64)

        # pandas 3 may expose read-only arrays due copy-on-write.
        runs, values = _coverage(
            coverage_df.Position.to_numpy(copy=True),
            coverage_df.Value.to_numpy(copy=True),
        )
        return Rle(runs, values)

    ranges = ranges.remove_strand() if not strand else ranges
    _kwargs = {
        "strand": strand,
        "value_col": value_col,
        "sparse": {"self": False},
    }  # already sparse
    kwargs.update(_kwargs)

    result = {
        k: _coverage_with_writable_arrays(v, value_col=kwargs.get("value_col"))
        for k, v in ranges.groupby(ranges.loc_columns)
    }

    if rpm:
        multiplier = 1e6 / len(ranges)
        result = {k: v * multiplier for k, v in result.items()}

    return PyRles(result)
